"""Convenience wrappers without direct OpenGL interaction.

@author: Stephan Wenger
@date: 2012-03-06
"""

from glitter.convenience.pipeline import *
from glitter.convenience.defaultpipeline import *
from glitter.convenience.copying import *
from glitter.convenience.mesh import *
from glitter.convenience.matrices import *

